#This file is part of Tryton.  The COPYRIGHT file at the top level of
#this repository contains the full copyright notices and license terms.
from datetime import date
from trytond.model import ModelView, fields
from trytond.pool import PoolMeta, Pool
from trytond.wizard import Wizard, StateTransition, StateView, Button
from trytond.transaction import Transaction

__all__ = ['AdvancePaymentAsk', 'AdvancePayment']
__metaclass__ = PoolMeta



class AdvancePaymentAsk(ModelView):
    'Advance Payment Ask'
    __name__ = 'account.invoice.advance_payment.ask'
    lines = fields.Many2Many('account.move.line', None, None,
            'Account Moves Lines')


class AdvancePayment(Wizard):
    'Advance Payment'
    __name__ = 'account.invoice.advance_payment'
    start_state = 'search_advance'
    search_advance = StateTransition()
    start = StateView('account.invoice.advance_payment.ask',
            'account_voucher.view_search_advance_payment_form', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Add', 'add_link', 'tryton-ok', default=True),
            ])
    add_link = StateTransition()

    @classmethod
    def __setup__(cls):
        super(AdvancePayment, cls).__setup__()
        cls._error_messages.update({
                'invalid_total_advance': 'Total payment advances can not '
                'be greater than Amount to Pay on invoice!',
            })

    def transition_search_advance(self):
        pool = Pool()
        Invoice = pool.get('account.invoice')
        Advance = pool.get('account.invoice.advance_payment.ask')
        invoice = Invoice(Transaction().context.get('active_id'))
        if invoice.type == 'in_invoice':
            account_types = ['receivable']
            debit_credit = ('debit', '>', 0)
        elif invoice.type == 'out_invoice':
            account_types = ['payable']
            debit_credit = ('credit', '>', 0)
        else:
            return 'end'
        domain = [
                ('reconciliation', '=', None),
                ('move.state', '=', 'posted'),
                ('party', '=', invoice.party.id),
                ('state', '=', 'valid'),
                ('account.kind', 'in', account_types),
        ]
        domain.append(debit_credit)
        Advance.lines.domain = domain
        return 'start'

    def transition_add_link(self):
        pool = Pool()
        Note = pool.get('account.note')
        Invoice = pool.get('account.invoice')
        MoveLine = pool.get('account.move.line')
        invoice = Invoice(Transaction().context.get('active_id'))
        configuration = pool.get('account.voucher_configuration')(1)

        move_lines = self.start.lines
        if not configuration.default_journal_note:
            Note.raise_user_error('missing_journal_note')

        lines_to_create = []
        accounts_to_reconcile = []
        lines_to_reconcile = []
        sum_debit = 0
        sum_credit = 0
        for line in move_lines:
            lines_to_create.append({
                'debit': line.credit,
                'credit': line.debit,
                'party': line.party.id,
                'account': line.account.id,
                'description': line.description,
            })
            sum_debit += line.debit
            sum_credit += line.credit
            accounts_to_reconcile.append(line.account.id)

        total_advance = abs(sum_debit - sum_credit)
        if total_advance > invoice.amount_to_pay:
            self.raise_user_error('invalid_total_advance')

        lines_to_create.append({
                'debit': sum_debit,
                'credit': sum_credit,
                'party': invoice.party.id,
                'account': invoice.account.id,
                'description': invoice.description,
        })

        note, = Note.create([{
            'description': '',
            'journal': configuration.default_journal_note.id,
            'date': date.today(),
            'state': 'draft',
            'lines': [('create', lines_to_create)],
        }])
        Note.post([note])
        lines_to_reconcile = move_lines
        payment_lines = []
        for line in note.move.lines:
            if line.account.id == invoice.account.id:
                payment_lines.append(line)
            if line.account.id in accounts_to_reconcile:
                lines_to_reconcile.append(line)

        Invoice.write([invoice], {
            'payment_lines': [('add', payment_lines)],
            })

        # TODO Reconcile advance
        if lines_to_reconcile:
            MoveLine.reconcile(lines_to_reconcile)
        return 'end'
