#This file is part of Tryton.  The COPYRIGHT file at the top level of
#this repository contains the full copyright notices and license terms.
from decimal import Decimal
from trytond.model import ModelView, ModelSQL, fields, Workflow
from trytond.wizard import Wizard, StateView, StateTransition, Button
from trytond.transaction import Transaction
from trytond.pyson import Eval, In, Or
from trytond.pool import Pool
from trytond.modules.company import CompanyReport
from trytond.report import Report

__all__ = ['VoucherPayMode', 'Voucher', 'VoucherLine', 'SelectLinesAsk',
    'SelectLines', 'VoucherReport', 'Note', 'NoteLine',
    'NoteReport', 'ChecksReport']

_STATES = {
    'readonly': Eval('state') != 'draft',
}

_STATES_NOTE = {
    'readonly': Eval('state') != 'draft',
}

TYPE_PAYMENTS = [
            ('', ''),
            ('check', 'Check'),
            ('cash', 'Cash'),
            ('card_terminal', 'Card Terminal'),
            ('transfer', 'Transfer'),
            ]

def getTranslated(obj, field):
    """
    Get the translation for selection field
    """
    selection = dict(obj.fields_get([field])[field]['selection'])
    return selection[getattr(obj, field)]


class Voucher(Workflow, ModelSQL, ModelView):
    'Voucher'
    __name__ = 'account.voucher'
    _rec_name = 'number'
    number = fields.Char('Number', readonly=True, help="Voucher Number")
    party = fields.Many2One('party.party', 'Party', required=True,
            states=_STATES, context={'party': Eval('party')})
    bank = fields.Many2One('bank', 'Bank', states={
            'readonly': In(Eval('state'), ['posted']),
            'invisible': Eval('payment_type') == 'cash',
            'required': In(Eval('payment_type'), ['check', 'card_terminal', 'transfer']),
            })
    check_number = fields.Char('Check Number', states={
            'readonly': In(Eval('state'), ['posted', 'cancel']),
            'invisible': Eval('payment_type') != 'check',
            'required': Eval('payment_type') == 'check',
            }, depends=['payment_mode'])
    payment_mode = fields.Many2One('account.voucher.paymode', 'Payment Mode', 
            select=True, states=_STATES)
    payment_type = fields.Char('Payment Type', depends=['payment_mode'])
    voucher_type = fields.Selection([
            ('payment', 'Payment'),
            ('receipt', 'Receipt'),
            ], 'Type', select=True, required=True, states=_STATES)
    date = fields.Date('Date', required=True, states=_STATES)
    journal = fields.Many2One('account.journal', 'Journal', required=True,
            depends=['voucher_type'], states=_STATES)
    currency = fields.Many2One('currency.currency', 'Currency', required=True,
            states=_STATES)
    company = fields.Many2One('company.company', 'Company', required=True,
            states=_STATES)
    lines = fields.One2Many('account.voucher.line', 'voucher', 'Lines',
            states=_STATES, context={'party': Eval('party')})
    comment = fields.Text('Comment', states=_STATES)
    description = fields.Char('Description', states=_STATES)
    state = fields.Selection([
            ('draft', 'Draft'),
            ('processed', 'Processed'),
            ('cancel', 'Cancel'),
            ('posted', 'Posted'),
            ], 'State', select=True, readonly=True)
    amount_to_pay = fields.Numeric('Amount To Pay', digits=(16, 2),
            select=True, states={'readonly': True})
    move = fields.Many2One('account.move', 'Move',
            states={'readonly': Or(
                    (Eval('state') != 'draft'),
                    (Eval('source') != 'move'),
                    ),
            })
    origin = fields.Reference('Origin', selection='get_origin', 
            select=True, depends=['state'], states={
                'readonly': Eval('state') != 'draft',
                })
    delivered = fields.Boolean('Delivered ?', states={
            'readonly': In(Eval('state'), ['posted', 'cancel']),
            'invisible': Eval('payment_type') != 'check',
            })
    account = fields.Many2One('account.account', 'Account', required=True,
        states=_STATES, domain=[
            ('company', '=', Eval('context', {}).get('company', -1)),
            ('kind', '!=', 'view'),
            ])

    @classmethod
    def __setup__(cls):
        super(Voucher, cls).__setup__()
        cls._error_messages.update({
                'missing_pay_lines': 'You have to enter lines to pay!',
                'delete_voucher': 'You can not delete a voucher that is posted!',
                'missing_journal_account': 'Missing debit or credit account on journal!',
                })
        cls._buttons.update({
                'draft': {
                    'invisible': Eval('state') == 'draft',
                    },
                'post': {
                    'invisible': Eval('state') != 'processed',
                    },
                'process': {
                    'invisible': Eval('state') != 'draft',
                    },
                'cancel': {
                    'invisible': Eval('state') == 'cancel',
                },
                'select_lines': {
                    'invisible': Eval('state') == 'posted',
                    },
                })
        cls._transitions |= set((
                ('draft', 'processed'),
                ('draft', 'cancel'),
                ('processed', 'draft'),
                ('processed', 'posted'),
                ('processed', 'cancel'),
                ('posted', 'cancel'),
                ('posted', 'draft'),
                ('cancel', 'draft'),
                ))
        cls._order.insert(0, ('date', 'DESC'))
        cls._order.insert(1, ('number', 'DESC'))

    @classmethod
    def __register__(cls, module_name):
        super(Voucher, cls).__register__(module_name)

    @staticmethod
    def default_state():
        return 'draft'

    @staticmethod
    def default_bank():
        VoucherConfig = Pool().get('account.voucher_configuration')
        configuration = VoucherConfig(1)
        if configuration.default_bank_payment:
            return configuration.default_bank_payment.id

    @staticmethod
    def default_currency():
        Company = Pool().get('company.company')
        company_id = Transaction().context.get('company')
        if company_id:
            return Company(company_id).currency.id

    @staticmethod
    def default_company():
        return Transaction().context.get('company')

    @staticmethod
    def default_date():
        Date = Pool().get('ir.date')
        return Date.today()

    @classmethod
    def copy(cls, vouchers, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['number'] = None
        default['state'] = 'draft'
        default['move'] = None
        default['lines'] = []
        default['check_number'] = '-'
        default['date'] = Pool().get('ir.date').today()
        new_vouchers = []
        for voucher in vouchers:
            new_voucher, = super(Voucher, cls).copy(
                    [voucher], default=default
                    )
            new_vouchers.append(new_voucher)
        return new_vouchers

    @fields.depends('payment_mode', 'journal', 'payment_type')
    def on_change_payment_mode(self):
        if not self.payment_mode:
            return {
                'payment_type': None,
                'bank': None,
                }
        changes = {}
        changes['journal'] = self.payment_mode.payment_journal.id
        changes['payment_type'] = self.payment_mode.payment_type
        changes['account'] = self.payment_mode.account.id
        if self.payment_mode.bank:
            changes['bank'] = self.payment_mode.bank.id
        return changes

    @fields.depends('lines', 'amount_to_pay')
    def on_change_lines(self):
        val = Decimal(0)
        if self.lines:
            val = self._get_amount_to_pay()
        changes = {'amount_to_pay': val}
        return changes

    """
    def create_lines_from_move_lines(self):
        values = []
        for mvline in self.moves_lines:
            if mvline.debit > 0 and self.voucher_type == 'payment':
                amount = mvline.debit
                line_type = 'debit'
            elif mvline.credit > 0 and self.voucher_type == 'receipt':
                line_type = 'credit'
                amount = mvline.credit
            else:
                continue
            values.append({
                'voucher': self.id,
                'account': mvline.account.id,
                'detail': mvline.description,
                'amount_original': None,
                'amount': amount,
                'move_line': mvline.id,
            })
        return values
    """

    def set_number(self):
        Sequence = Pool().get('ir.sequence')
        AccountVoucherSequence = Pool().get('account.voucher.sequence')
        sequence = AccountVoucherSequence(1)
        if self.number:
            return
        if self.voucher_type == 'receipt':
            seq = sequence.voucher_receipt_sequence.id
        else:
            seq = sequence.voucher_payment_sequence.id
        self.write([self], {'number': Sequence.get_id(seq)})

    @classmethod
    @ModelView.button
    @Workflow.transition('processed')
    def process(cls, vouchers):
        for voucher in vouchers:
            if not voucher.journal.debit_account or not voucher.journal.credit_account:
                voucher.raise_user_error('missing_journal_account')
            if voucher.amount_to_pay <= Decimal("0.0"):
                voucher.raise_user_error('missing_pay_lines')
            voucher.set_number()
            if not voucher.move:
                voucher.create_move()

    @classmethod
    @ModelView.button
    @Workflow.transition('draft')
    def draft(cls, vouchers):
        pass

    @classmethod
    @ModelView.button
    @Workflow.transition('cancel')
    def cancel(cls, vouchers):
        Move = Pool().get('account.move')
        for voucher in vouchers:
            if voucher.move:
                Move.delete([voucher.move])

    @classmethod
    @ModelView.button
    @Workflow.transition('posted')
    def post(cls, vouchers):
        for voucher in vouchers:
            voucher._post_move()

    @classmethod
    @ModelView.button_action('account_voucher.wizard_select_lines')
    def select_lines(cls, vouchers):
        pass

    @staticmethod
    def _get_origin():
        'Return list of Model names for origin Reference'
        return []

    @classmethod
    def get_origin(cls):
        Model = Pool().get('ir.model')
        models = cls._get_origin()
        models = Model.search([
                ('model', 'in', models),
        ])
        return [(None, '')] + [(m.model, m.name) for m in models]

    def get_payment_type(self, name=None):
        if self.payment_mode:
            return self.payment_mode.payment_type

    def _get_amount_to_pay(self):
        res = Decimal('0.0')
        if self.lines:
            for line in self.lines:
                if line.amount:
                    res += line.amount
        return res

    def create_move(self):
        pool = Pool()
        Move = pool.get('account.move')
        Period = pool.get('account.period')
        MoveLine = pool.get('account.move.line')
        move_lines = []
        line_move_ids = []

        # Check amount
        period_id = Period.find(self.company.id, date=self.date)

        move, = Move.create([{
                'journal': self.journal.id,
                'period': period_id,
                'date': self.date,
                'origin': str(self),
                'state': 'draft',
                'description': self.description,
        }])

        self.write([self], {
                'move': move.id,
        })

        credit = Decimal('0.0')
        debit = Decimal('0.0')
        if self.voucher_type == 'receipt':
            debit = self.amount_to_pay
        else:
            credit = self.amount_to_pay
        move_lines.append({
            'description': self.description,
            'account': self.account.id,
            'debit': debit,
            'credit': credit,
            'move': move.id,
            'journal': self.journal.id,
            'period': period_id,
            'party': self.party.id,
        })

        total = self.amount_to_pay

        for line in self.lines:
            if not line.amount:
                continue
            credit = Decimal('0.0')
            debit = Decimal('0.0')
            line_move_ids.append(line.move_line)
            if self.voucher_type == 'receipt':
                credit = line.amount
            else:
                debit = line.amount
            total -= line.amount
            description = ''
            if line.move_line and line.move_line.origin:
                if hasattr(line.move_line.origin, 'number'):
                    description = line.move_line.origin.number
                else:
                    description = line.move_line.origin.rec_name
            elif line.detail:
                description = line.detail
            move_lines.append({
                'description': description,
                'debit': debit,
                'credit': credit,
                'account': line.account.id,
                'move': move.id,
                'journal': self.journal.id,
                'period': period_id,
                'date': self.date,
                'party': self.party.id,
            })
        MoveLine.create(move_lines)

    def _post_move(self):
        pool = Pool()
        Invoice = pool.get('account.invoice')
        Move = pool.get('account.move')
        MoveLine = pool.get('account.move.line')
        Move.post([self.move])

        # reconcile check
        reconcile_lines = []
        for line in self.lines:
            if line.amount == Decimal('0.00') or not line.move_line:
                continue

            for move_line in self.move.lines:
                if self.voucher_type == 'payment':
                    if move_line.debit != line.amount or move_line.reconciliation:
                        continue
                else:
                    if move_line.credit != line.amount:
                        continue
                reconcile_lines.append(move_line)
                if line.move_line.origin and \
                    line.move_line.origin.__name__ == 'account.invoice':
                    invoice = Invoice(line.move_line.origin.id)
                    if invoice.type in ('in_invoice', 'out_credit_note'):
                        amount = -(line.amount)
                    else:
                        amount = line.amount
                    invoice_reconcile_lines, remainder = invoice.get_reconcile_lines_for_amount(amount)
                    Invoice.write([invoice], {
                        'payment_lines': [('add', [move_line.id])],
                    })
                    if remainder == Decimal('0.00'):
                        reconcile_lines.extend(invoice_reconcile_lines)
                else:
                    reconcile_lines.append(line.move_line)
        balance = 0
        for rl in reconcile_lines:
            balance += rl.debit - rl.credit
        if reconcile_lines and balance == 0:
            MoveLine.reconcile(reconcile_lines)

    @classmethod
    def delete(cls, vouchers):
        pool = Pool()
        Line = pool.get('account.voucher.line')
        Move = pool.get('account.move')
        for voucher in vouchers:
            if voucher.state == 'posted':
                cls.raise_user_error('delete_voucher')
            if voucher.move:
                Move.delete([voucher.move])
        Line.delete([l for v in vouchers for l in v.lines])
        return super(Voucher, cls).delete(vouchers)


class VoucherLine(ModelSQL, ModelView):
    'Voucher Line'
    __name__ = 'account.voucher.line'
    voucher = fields.Many2One('account.voucher', 'Voucher', required=True)
    detail = fields.Char('Detail')
    cost_center = fields.Char('Cost Center')
    account = fields.Many2One('account.account', 'Account', required=True,
        domain=[
            ('company', '=', Eval('context', {}).get('company', -1)),
            ('kind', '!=', 'view'),
            ])
    amount = fields.Numeric('Amount', digits=(16, 2), required=True)
    move_line = fields.Many2One('account.move.line', 'Move Line',
            domain=[
                ('party', '=', Eval('_parent_voucher', {}).get('party')),
                ('move.state', '=', 'posted'),
                ('state', '=', 'valid'),
                ('account.kind', 'in', ['payable', 'receivable']),
            ])
    amount_original = fields.Numeric('Original Amount', digits=(16, 2),
            readonly=True)

    def get_reference(self, name):
        Invoice = Pool().get('account.invoice')
        if self.move_line.move:
            invoices = Invoice.search(
                [('move', '=', self.move_line.move.id)])
            if invoices:
                return invoices[0].reference

    @fields.depends('move_line', 'account', 'amount', 'detail')
    def on_change_move_line(self):
        changes = {}
        if not self.move_line:
            return {}
        if self.move_line.debit > 0:
            amount = self.move_line.debit
        elif self.move_line.credit > 0:
            amount = self.move_line.credit
        else:
            amount = Decimal(0)
        changes = {
            'amount': amount,
            'account': self.move_line.account.id,
            'detail': self.move_line.description,
        }
        return changes


class SelectLinesAsk(ModelView):
    'Select Lines Assistant'
    __name__ = 'account.voucher.select_lines.ask'
    lines = fields.Many2Many('account.move.line', None, None,
            'Account Moves Lines')


class SelectLines(Wizard):
    'Select Lines'
    __name__ = 'account.voucher.select_lines'
    start_state = 'search_lines'
    search_lines = StateTransition()
    start = StateView('account.voucher.select_lines.ask',
            'account_voucher.view_search_lines_form', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Add', 'add_lines', 'tryton-ok', default=True),
            ])
    add_lines = StateTransition()

    def transition_search_lines(self):
        pool = Pool()
        Voucher = pool.get('account.voucher')
        Select = pool.get('account.voucher.select_lines.ask')
        voucher = Voucher(Transaction().context.get('active_id'))
        if voucher.voucher_type == 'receipt':
            account_types = ['receivable']
            debit_credit = ('debit', '>', 0)
        else:
            account_types = ['payable']
            debit_credit = ('credit', '>', 0)
        line_domain = [
                ('account.kind', 'in', account_types),
                ('state', '=', 'valid'),
                ('reconciliation', '=', None),
                ('move.state', '=', 'posted'),
                ('party', '=', voucher.party.id),
        ]
        line_domain.append(debit_credit)
        Select.lines.domain = line_domain
        return 'start'

    def transition_add_lines(self):
        pool = Pool()
        Voucher = pool.get('account.voucher')
        Line = pool.get('account.voucher.line')
        Invoice = pool.get('account.invoice')
        voucher = Voucher(Transaction().context.get('active_id'))
        
        Model = Pool().get('ir.model')
        move_lines_ids = self.start.lines
        sum_amount_to_pay = Decimal('0')
        lines_to_create = []
        for line in move_lines_ids:
            if line.credit > Decimal('0'):
                amount = line.credit
            else:
                amount = line.debit
            model = None
            amount_original = amount
            if line.origin and line.origin.__name__ == 'account.invoice':
                model, = Model.search([
                        ('model', '=', line.origin.__name__),
                        ])
                amount_to_pay = Invoice.get_amount_to_pay(
                        [line.origin], 'amount_to_pay'
                        )
                amount_to_pay = amount_to_pay[line.origin.id]
                detail = (model.name + ' ' + line.origin.number)
                if amount_to_pay < amount:
                    amount = amount_to_pay
            else:
                amount_to_pay = amount
                detail = line.description
            lines_to_create.append({
                'voucher': voucher.id,
                'detail': detail,
                'account': line.account.id,
                'amount_original': amount_original,
                'amount': amount,
                'move_line': line.id,
            })
            sum_amount_to_pay += amount
        Line.create(lines_to_create)

        voucher.amount_to_pay = sum_amount_to_pay
        voucher.save()
        return 'end'


class VoucherReport(Report):
    'Voucher Report'
    __name__ = 'account.voucher.report'

    @classmethod
    def parse(cls, report, objects, data, localcontext=None):
        Company = Pool().get('company.company')
        company_id = Transaction().context.get('company')
        localcontext['company'] = Company(company_id)
        conversor = None
        try:
            from numword import numword_es
            conversor = numword_es.NumWordES()
        except:
            print("Warning: Does not possible import numword module!")
            print("Please install it...!")

        new_objs = []
        for obj in objects:
            num = ''
            if obj.amount_to_pay and conversor:
                num = (conversor.cardinal(int(obj.amount_to_pay))).upper()
            setattr(obj, 'payword', num)
            new_objs.append(obj)
        return super(VoucherReport, cls).parse(report,
                new_objs, data, localcontext)


class VoucherPayMode(ModelSQL, ModelView):
    'Voucher Pay Mode'
    __name__ = 'account.voucher.paymode'
    name = fields.Char('Name', required=True)
    account = fields.Many2One('account.account', 'Account',
            required=True, domain=[('kind', '!=', 'view')])
    payment_type = fields.Selection(TYPE_PAYMENTS, 'Payment Type', 
            select=True)
    bank = fields.Many2One('bank', 'Bank', states={
            'invisible': Eval('payment_type') == 'cash',
            })
    bank_account = fields.Many2One('bank.account', 'Bank Account',
            states={'invisible': Eval('payment_type') == 'cash'})
    payment_journal = fields.Many2One('account.journal', 'Payment Journal',
            required=True)
    kind = fields.Selection([
            ('payment', 'Payments'),
            ('receipt', 'Receipts'),
            ('both', 'Both'),
            ], 'Kind', select=True, required=True,
            translate=True, states=_STATES)


class Note(Workflow, ModelSQL, ModelView):
    'Note'
    __name__ = 'account.note'
    _rec_name = 'number'
    number = fields.Char('Number', readonly=True, help="Voucher Number")
    description = fields.Char('Description', states=_STATES_NOTE)
    date = fields.Date('Date', required=True, states=_STATES_NOTE)
    journal = fields.Many2One('account.journal', 'Journal', required=True,
            states=_STATES_NOTE)
    currency = fields.Many2One('currency.currency', 'Currency', 
            required=True, states=_STATES_NOTE)
    company = fields.Many2One('company.company', 'Company', 
            required=True, states=_STATES_NOTE)
    lines = fields.One2Many('account.note.line', 'note', 'Lines',
            states=_STATES)
    state = fields.Selection([
            ('draft', 'Draft'),
            ('cancel', 'Cancel'),
            ('posted', 'Posted'),
            ], 'State', select=True, readonly=True, translate=True)
    move = fields.Many2One('account.move', 'Move', readonly=True)
    origin = fields.Reference('Origin', selection='get_origin', 
            select=True, states=_STATES_NOTE, depends=['state'])

    @classmethod
    def __setup__(cls):
        super(Note, cls).__setup__()
        cls._error_messages.update({
                'post_empty_note': ('You can not post note "%s" because it is '
                    'empty.'),
                'post_unbalanced_note': ('You can not post note "%s" because '
                    'it is an unbalanced.'),
                'missing_lines': 'You have to enter lines!',
                'delete_note': 'You can not delete a note that is posted!',
                'missing_journal_note': 'Missing journal note!',
                })
        cls._buttons.update({
                'draft': {
                    'invisible': Eval('state') != 'cancel',
                    },
                'post': {
                    'invisible': Eval('state') != 'draft',
                    },
                'cancel': {
                    'invisible': Eval('state') != 'draft',
                },
                })
        cls._transitions |= set((
                ('draft', 'cancel'),
                ('posted', 'cancel'),
                ('draft', 'posted'),
                ('cancel', 'draft'),
                ))
        cls._order.insert(0, ('date', 'DESC'))
        cls._order.insert(1, ('number', 'DESC'))

    @classmethod
    def __register__(cls, module_name):
        super(Note, cls).__register__(module_name)

    @staticmethod
    def default_state():
        return 'draft'

    @staticmethod
    def default_currency():
        Company = Pool().get('company.company')
        company_id = Transaction().context.get('company')
        if company_id:
            return Company(company_id).currency.id

    @staticmethod
    def default_company():
        return Transaction().context.get('company')

    @staticmethod
    def default_date():
        Date = Pool().get('ir.date')
        return Date.today()

    @staticmethod
    def default_journal():
        configuration = Pool().get('account.voucher_configuration')(1)
        if configuration.default_journal_note:
            return configuration.default_journal_note.id

    @classmethod
    def copy(cls, notes, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['number'] = None
        default['state'] = cls.default_state()
        default['lines'] = []

        new_notes = []
        for note in notes:
            new_note, = super(Note, cls).copy(
                    [note], default=default
                    )
            new_notes.append(new_note)
        return new_notes

    @classmethod
    @ModelView.button
    @Workflow.transition('draft')
    def draft(cls, vouchers):
        pass

    @classmethod
    @ModelView.button
    @Workflow.transition('posted')
    def post(cls, notes):
        for note in notes:
            cls._post_note(note)
            note.create_moves()
            note.set_number()

    def set_number(self):
        if self.number:
            return
        Sequence = Pool().get('ir.sequence')
        AccountVoucherSequence = Pool().get('account.voucher.sequence')
        sequence = AccountVoucherSequence(1)
        seq = sequence.voucher_notes_sequence.id
        self.write([self], {'number': Sequence.get_id(seq)})

    @classmethod
    def _post_note(cls, note):
        values = {'state': 'posted'}
        amount = Decimal('0.0')
        if not note.lines:
            cls.raise_user_error('post_empty_note', (note.rec_name,))
        company = None
        for line in note.lines:
            amount += line.debit - line.credit
            if not company:
                company = line.account.company
        if not company.currency.is_zero(amount):
            cls.raise_user_error('post_unbalanced_note', (note.rec_name,))
        cls.write([note], values)

    def create_moves(self):
        pool = Pool()
        Move = pool.get('account.move')
        MoveLine = pool.get('account.move.line')
        Period = pool.get('account.period')
        move_lines = []

        if not self.lines:
            self.raise_user_error('missing_lines')

        period_id = Period.find(self.company.id, date=self.date)
        move, = Move.create([{
                'journal': self.journal.id,
                'period': period_id,
                'date': self.date,
                'origin': str(self),
                'state': 'draft',
                'description': self.description,
                }])

        self.write([self], {
                'move': move.id,
                })

        for line in self.lines:
            move_lines.append({
                'description': line.description,
                'debit': line.debit,
                'credit': line.credit,
                'account': line.account.id,
                'move': move.id,
                'journal': self.journal.id,
                'period': period_id,
                'date': self.date,
                'party': line.party.id,
            })
        MoveLine.create(move_lines)
        Move.post([move])

    @staticmethod
    def _get_origin():
        'Return list of Model names for origin Reference'
        return []

    @classmethod
    def get_origin(cls):
        Model = Pool().get('ir.model')
        models = cls._get_origin()
        models = Model.search([
                ('model', 'in', models),
                ])
        return [(None, '')] + [(m.model, m.name) for m in models]


class NoteLine(ModelSQL, ModelView):
    'Note Line'
    __name__ = 'account.note.line'
    note = fields.Many2One('account.note', 'Note', select=True, required=True)
    debit = fields.Numeric('Debit', digits=(16, Eval('currency_digits', 2)),
        required=True, depends=['currency_digits', 'credit'])
    credit = fields.Numeric('Credit', digits=(16, Eval('currency_digits', 2)),
        required=True, depends=['currency_digits', 'debit'])
    account = fields.Many2One('account.account', 'Account', select=True, 
            domain=[('kind', '!=', 'view')], required=True)
    description = fields.Char('Description')
    party = fields.Many2One('party.party', 'Party', required=True,
                select=True)
    currency_digits = fields.Function(fields.Integer('Currency Digits'),
            'get_currency_digits')
    second_currency_digits = fields.Function(fields.Integer(
        'Second Currency Digits'), 'get_currency_digits')

    @classmethod
    def __setup__(cls):
        super(NoteLine, cls).__setup__()
        cls._order[0] = ('id', 'DESC')

    @staticmethod
    def default_currency_digits():
        return 2

    @staticmethod
    def default_debit():
        return Decimal(0)

    @staticmethod
    def default_credit():
        return Decimal(0)

    @staticmethod
    def default_party():
        return Transaction().context.get('company')

    @fields.depends('account', 'debit', 'credit', 'note')
    def on_change_debit(self):
        changes = {}
        if self.debit:
            changes['credit'] = Decimal('0.0')
        return changes

    @fields.depends('account', 'debit', 'credit', 'note')
    def on_change_credit(self):
        changes = {}
        if self.credit:
            changes['debit'] = Decimal('0.0')
        return changes

    @fields.depends('account', 'debit', 'credit', 'note')
    def on_change_account(self):
        changes = {}
        if self.account:
            changes['currency_digits'] = self.account.currency_digits
            if self.account.second_currency:
                changes['second_currency_digits'] = \
                    self.account.second_currency.digits
        return changes

    @classmethod
    def delete(cls, lines):
        NoteLine = Pool().get('account.note')
        super(NoteLine, cls).delete(lines)

    @classmethod
    def copy(cls, lines, default=None):
        if default is None:
            default = {}
        if 'note' not in default:
            default['note'] = None
        return super(NoteLine, cls).copy(lines, default=default)

    @classmethod
    def get_currency_digits(cls, lines, names):
        digits = {}
        for line in lines:
            for name in names:
                digits.setdefault(name, {})
                digits[name].setdefault(line.id, 2)
                if name == 'currency_digits':
                    digits[name][line.id] = line.account.currency_digits
                elif name == 'second_currency_digits':
                    second_currency = line.account.second_currency
                    if second_currency:
                        digits[name][line.id] = second_currency.digits
        return digits


class NoteReport(CompanyReport):
    'Account Note Report'
    __name__ = 'account.note.report'

    @classmethod
    def parse(cls, report, objects, data, localcontext=None):
        for obj in objects:
            sum_credit = 0
            sum_debit = 0
            for line in obj.lines:
                sum_credit += line.credit
                sum_debit += line.debit
        data['sum_credit'] = sum_credit
        data['sum_debit'] = sum_debit
        return super(NoteReport, cls).parse(report,
                objects, data, localcontext)


class ChecksReport(Report):
    'Sheet Checks Control'
    __name__ = 'account.voucher.checks_report'

    @classmethod
    def parse(cls, report, objects, data, localcontext=None):
        user = Pool().get('res.user')(Transaction().user)
        localcontext['company'] = user.company
        vouchers = [obj for obj in objects if obj.number and obj.payment_type == 'check']
        objects = []
        for obj in vouchers:
            setattr(obj, 'state', getTranslated(obj, 'state'))
            if obj.delivered:
                x = 'X'
            else:
                x = ''
            setattr(obj, 'is_delivered', x)
            objects.append(obj)
        return super(ChecksReport, cls).parse(report,
                objects, data, localcontext)
